import * as cdk from "aws-cdk-lib";
import * as gamelift from "aws-cdk-lib/aws-gamelift";
import * as lambda_nodejs from "aws-cdk-lib/aws-lambda-nodejs";
import * as apigateway from "aws-cdk-lib/aws-apigateway";
import * as iam from "aws-cdk-lib/aws-iam";

import { type Construct } from "constructs";
import { NagSuppressions } from "cdk-nag";

export class CdkStack extends cdk.Stack {
  constructor(scope: Construct, id: string, props?: cdk.StackProps) {
    super(scope, id, props);

    NagSuppressions.addStackSuppressions(this, [
      {
        id: "AwsSolutions-IAM4",
        reason:
          "AWSLambdaBasicExecutionRole is a default role attached by AWS Lambda.",
      },
      {
        id: "AwsSolutions-IAM5",
        reason:
          "CreateGameSession and DescribeGameSessions does not allow to specify resources.",
      },
      {
        id: "AwsSolutions-APIG1",
        reason: "This sample is used only for dev or sample, not for prod.",
      },
      {
        id: "AwsSolutions-APIG2",
        reason:
          "Backend Lambda function does not use any parameters from user input.",
      },
      {
        id: "AwsSolutions-APIG4",
        reason:
          "This sample is used only for dev or sample, not for prod. And pre-defined api-key is required.",
      },
      {
        id: "AwsSolutions-COG4",
        reason:
          "This sample is used only for dev or sample, not for prod. And pre-defined api-key is required.",
      },
      {
        id: "AwsSolutions-APIG6",
        reason: "This sample is used only for dev or sample, not for prod.",
      },
    ]);

    // GameLift ロケーションを定義
    const customLocation = new gamelift.CfnLocation(this, "customLocation", {
      locationName: "custom-localMachine",
    });

    // GameLift ロケーションを含む GameLift Anywhere フリートを定義
    const anywhereFleet = new gamelift.CfnFleet(this, "anywhereFleet", {
      name: "builders-flash-sample-game-anywhere",
      computeType: "ANYWHERE",
      anywhereConfiguration: {
        cost: "0.0001",
      },
      locations: [{ location: customLocation.locationName }],
    });
    anywhereFleet.addDependency(customLocation);

    // 使用可能な GameSession を検索する Lambda 関数を定義
    const listAvailableGameSessions = new lambda_nodejs.NodejsFunction(
      this,
      "listAvailableGameSessions",
      {
        entry: "./src/listAvailableGameSessions.ts",
        handler: "lambdaHandler",
        timeout: cdk.Duration.seconds(10),
        memorySize: 1024,
        runtime: cdk.aws_lambda.Runtime.NODEJS_18_X,
      },
    );
    // 定義した Lambda 関数に環境変数を指定し、必要な権限を付与
    listAvailableGameSessions.addEnvironment(
      "FLEETID",
      anywhereFleet.attrFleetId,
    );
    listAvailableGameSessions.addEnvironment(
      "LOCATIONNAME",
      customLocation.locationName,
    );
    listAvailableGameSessions.addToRolePolicy(
      new iam.PolicyStatement({
        actions: ["gamelift:DescribeGameSessions"],
        resources: ["*"],
      }),
    );

    // 新しく GameSession を作成する Lambda 関数を定義
    const createGameSession = new lambda_nodejs.NodejsFunction(
      this,
      "createGameSession",
      {
        entry: "./src/createGameSession.ts",
        handler: "lambdaHandler",
        timeout: cdk.Duration.seconds(10),
        memorySize: 1024,
        runtime: cdk.aws_lambda.Runtime.NODEJS_18_X,
      },
    );
    // 定義した Lambda 関数に環境変数を指定し、必要な権限を付与
    createGameSession.addEnvironment("FLEETID", anywhereFleet.attrFleetId);
    createGameSession.addEnvironment(
      "LOCATIONNAME",
      customLocation.locationName,
    );
    createGameSession.addToRolePolicy(
      new iam.PolicyStatement({
        actions: ["gamelift:CreateGameSession"],
        resources: ["*"],
      }),
    );

    // API Gateway REST API を定義
    const api = new apigateway.RestApi(this, "Api", {
      restApiName: "BackendServiceApi",
    });
    const defaultroot = api.root;

    // REST API に Lambda 関数を紐づける Resource を定義
    const resourcePlayerSession = defaultroot.addResource("gamesession");
    resourcePlayerSession.addMethod(
      "PUT",
      new apigateway.LambdaIntegration(createGameSession),
      { apiKeyRequired: true },
    );
    const resourceGameSessions = defaultroot.addResource("gamesessions");
    resourceGameSessions.addMethod(
      "GET",
      new apigateway.LambdaIntegration(listAvailableGameSessions),
      { apiKeyRequired: true },
    );

    // REST API に対し API Key を定義
    const apikey = api.addApiKey("clientKey", {
      apiKeyName: "samplebuildersflashkey"
    });
    const usagePlan = api.addUsagePlan("usagePlan", {});
    usagePlan.addApiKey(apikey);
    usagePlan.addApiStage({ stage: api.deploymentStage });

    // 定義したリソースを出力
    new cdk.CfnOutput(this, "APIDomain", {
      value: api.url,
      description: "",
      exportName: "APIDomain",
    });
    new cdk.CfnOutput(this, "FleetID", {
      value: anywhereFleet.attrFleetId,
      description: "",
      exportName: "GameLiftFleetID",
    });
    new cdk.CfnOutput(this, "LocationName", {
      value: customLocation.locationName,
      description: "",
      exportName: "GameLiftFleetLocationName",
    });
  }
}
