package main

import (
	"flag"
	"log"
	"net/http"
	"strconv"
	"sync"
	"time"

	"github.com/google/uuid"
	"github.com/gorilla/websocket"
)

var upgrader = websocket.Upgrader{}

// Message チャンネルで送受信するデータの型を定義
type Message struct {
	Type int
	Body []byte
}

// Client 接続したクライアントごとのコネクションと未送信メッセージキューを定義
type Client struct {
	Conn          *websocket.Conn
	SendQueueChan chan Message
}

// Clients 接続したクライアントの情報をもつ Client を Key にもつ Map
var Clients = make(map[*Client]bool)
var Broadcast = make(chan Message)
var Mutex = &sync.Mutex{}

// メッセージの受信を確認し、受信したらブロードキャスト用chに入れる
func listenMessageAsync(client *Client) {
	defer close(client.SendQueueChan)
	for {
		mt, message, err := client.Conn.ReadMessage()
		log.Println("[WebSocketServer]read: ", message)

		if err != nil {
			log.Println("[WebSocketServer]read: ", err)
			disconnectWebSocket(client)
			break
		}
		Broadcast <- Message{mt, message}
	}
}

// 送信 チャンネル のメッセージを WebSocket でクライアントに送信
func processSendQueueAsync(client *Client) {
	for {
		msg, ok := <-client.SendQueueChan
		if !ok {
			return
		}
		err := client.Conn.WriteMessage(msg.Type, msg.Body)
		if err != nil {
			log.Println("[WebSocketServer]WriteMessage failed: ", err.Error())
			disconnectWebSocket(client)
			break
		}
	}
}

// チャンネルに溜まったメッセージをブロードキャストするため各クライアントのチャンネルに配信.
func broadcastMessagesAsync() {
	for {
		msg := <-Broadcast
		for client := range Clients {
			client.SendQueueChan <- msg
		}
	}
}

func disconnectWebSocket(client *Client) {
	Mutex.Lock()
	defer Mutex.Unlock()
	client.Conn.Close()
	delete(Clients, client)
}

// HTTP リクエストを受け、WebSocket 接続を開始する
func connect(w http.ResponseWriter, r *http.Request) {
	conn, err := upgrader.Upgrade(w, r, nil)
	if err != nil {
		log.Println("[WebSocketServer]conn: ", err)
		return
	}

	client := Client{Conn: conn, SendQueueChan: make(chan Message)}
	Mutex.Lock()
	defer Mutex.Unlock()
	Clients[&client] = true

	// メッセージを受信する goroutine を作成
	go listenMessageAsync(&client)
	// 自身のchに溜まったメッセージを送信する goroutine を作成
	go processSendQueueAsync(&client)
}

// HTTP リクエストに応答する
func healthcheck(w http.ResponseWriter, r *http.Request) {
	w.Write([]byte("Status is green!\n"))
	log.Println("[Receive HTTP request] Status is green ")
}

func main() {
	// HTTP リクエストをリッスンする
	http.HandleFunc("/connect", connect)
	http.HandleFunc("/healthcheck", healthcheck)
	http.HandleFunc("/", healthcheck)

	// メッセージをブロードキャストするための goroutine を作成
	go broadcastMessagesAsync()

	// 環境変数もしくは固定の値を取得する
	processUUID, _ := uuid.NewUUID()
	var websocketurl = flag.String("websocketurl", "wss://ap-northeast-1.api.amazongamelift.com", "WebSocket URL for sync gamelift status")
	var processid = flag.String("processid", "processid"+processUUID.String(), "Random Process ID")
	var hostid = flag.String("hostid", "", "Compute name with RegisterCompute API")
	var fleetid = flag.String("fleetid", "", "")
	var authtoken = flag.String("authtoken", "", "")
	var port = flag.Int("port", 80, "")
	flag.Parse()

	// GameLiftService と同期のために GameLift ServerSDK を init する
	gameliftServerSDKSetup(*websocketurl, *processid, *hostid, *fleetid, *authtoken, *port)
	defer gameliftServerSDKDestroy()

	// HTTP サーバーを起動し、80 番ポートをリッスンする
	srv := &http.Server{
		Addr:              ":" + strconv.Itoa(*port),
		Handler:           nil,
		ReadHeaderTimeout: 20 * time.Second,
	}
	if err := srv.ListenAndServe(); err != nil {
		log.Panicln("[WebSocketServer]Serve Error:", err)
	}
}
